import click
from py42.clients.detectionlists import RiskTags
from py42.services.detectionlists.high_risk_employee import HighRiskEmployeeFilters

from code42cli.bulk import generate_template_cmd_factory
from code42cli.bulk import run_bulk_process
from code42cli.click_ext.groups import OrderedGroup
from code42cli.cmds.detectionlists import add_risk_tags as _add_risk_tags
from code42cli.cmds.detectionlists import ALL_FILTER
from code42cli.cmds.detectionlists import get_choices
from code42cli.cmds.detectionlists import handle_filter_choice
from code42cli.cmds.detectionlists import handle_list_args
from code42cli.cmds.detectionlists import list_employees
from code42cli.cmds.detectionlists import remove_risk_tags as _remove_risk_tags
from code42cli.cmds.detectionlists import update_user
from code42cli.cmds.detectionlists.options import cloud_alias_option
from code42cli.cmds.detectionlists.options import notes_option
from code42cli.cmds.detectionlists.options import username_arg
from code42cli.cmds.shared import get_user_id
from code42cli.file_readers import read_csv_arg
from code42cli.options import format_option
from code42cli.options import sdk_options
from code42cli.util import deprecation_warning

DEPRECATION_TEXT = "(DEPRECATED): Use `code42 watchlists` commands instead."


def _get_filter_choices():
    filters = HighRiskEmployeeFilters.choices()
    return get_choices(filters)


filter_option = click.option(
    "--filter",
    help=f"High risk employee filter options. Defaults to {ALL_FILTER}.",
    type=click.Choice(_get_filter_choices()),
    default=ALL_FILTER,
    callback=lambda ctx, param, arg: handle_filter_choice(arg),
)


risk_tag_option = click.option(
    "-t",
    "--risk-tag",
    multiple=True,
    type=click.Choice(RiskTags.choices()),
    help="Risk tags associated with the employee.",
)


@click.group(
    cls=OrderedGroup,
    help=f"{DEPRECATION_TEXT}\n\nAdd and remove employees from the High Risk Employees detection list.",
)
@sdk_options(hidden=True)
def high_risk_employee(state):
    pass


@high_risk_employee.command(
    "list",
    help=f"{DEPRECATION_TEXT}\n\nLists the employees on the High Risk Employee list.",
)
@sdk_options()
@format_option
@filter_option
def _list(state, format, filter):
    deprecation_warning(DEPRECATION_TEXT)
    employee_generator = _get_high_risk_employees(state.sdk, filter)
    list_employees(employee_generator, format)


@high_risk_employee.command(
    help=f"{DEPRECATION_TEXT}\n\nAdd a user to the high risk employees detection list."
)
@cloud_alias_option
@notes_option
@risk_tag_option
@username_arg
@sdk_options()
def add(state, username, cloud_alias, risk_tag, notes):
    deprecation_warning(DEPRECATION_TEXT)
    _add_high_risk_employee(state.sdk, username, cloud_alias, risk_tag, notes)


@high_risk_employee.command(
    help=f"{DEPRECATION_TEXT}\n\nRemove a user from the high risk employees detection list."
)
@username_arg
@sdk_options()
def remove(state, username):
    deprecation_warning(DEPRECATION_TEXT)
    _remove_high_risk_employee(state.sdk, username)


@high_risk_employee.command(
    help=f"{DEPRECATION_TEXT}\n\nAssociates risk tags with a user."
)
@username_arg
@risk_tag_option
@sdk_options()
def add_risk_tags(state, username, risk_tag):
    deprecation_warning(DEPRECATION_TEXT)
    _add_risk_tags(state.sdk, username, risk_tag)


@high_risk_employee.command(
    help=f"{DEPRECATION_TEXT}\n\nDisassociates risk tags from a user."
)
@username_arg
@risk_tag_option
@sdk_options()
def remove_risk_tags(state, username, risk_tag):
    deprecation_warning(DEPRECATION_TEXT)
    _remove_risk_tags(state.sdk, username, risk_tag)


@high_risk_employee.group(
    cls=OrderedGroup,
    help=f"{DEPRECATION_TEXT}\n\nTools for executing high risk employee actions in bulk.",
)
@sdk_options(hidden=True)
def bulk(state):
    pass


HIGH_RISK_EMPLOYEE_CSV_HEADERS = ["username", "cloud_alias", "risk_tag", "notes"]
RISK_TAG_CSV_HEADERS = ["username", "tag"]
REMOVE_EMPLOYEE_HEADERS = ["username"]

high_risk_employee_generate_template = generate_template_cmd_factory(
    group_name="high_risk_employee",
    commands_dict={
        "add": HIGH_RISK_EMPLOYEE_CSV_HEADERS,
        "remove": REMOVE_EMPLOYEE_HEADERS,
        "add-risk-tags": RISK_TAG_CSV_HEADERS,
        "remove-risk-tags": RISK_TAG_CSV_HEADERS,
    },
)
bulk.add_command(high_risk_employee_generate_template)


@bulk.command(
    name="add",
    help=f"{DEPRECATION_TEXT}\n\nBulk add users to the high risk employees detection list using a "
    f"CSV file with format: {','.join(HIGH_RISK_EMPLOYEE_CSV_HEADERS)}.",
)
@read_csv_arg(headers=HIGH_RISK_EMPLOYEE_CSV_HEADERS)
@sdk_options()
def bulk_add(state, csv_rows):
    deprecation_warning(DEPRECATION_TEXT)
    sdk = state.sdk

    def handle_row(username, cloud_alias, risk_tag, notes):
        _add_high_risk_employee(sdk, username, cloud_alias, risk_tag, notes)

    run_bulk_process(
        handle_row,
        csv_rows,
        progress_label="Adding users to high risk employee detection list:",
    )


@bulk.command(
    name="remove",
    help=f"{DEPRECATION_TEXT}\n\nBulk remove users from the high risk employees detection list "
    f"using a CSV file with format {','.join(REMOVE_EMPLOYEE_HEADERS)}.",
)
@read_csv_arg(headers=REMOVE_EMPLOYEE_HEADERS)
@sdk_options()
def bulk_remove(state, csv_rows):
    deprecation_warning(DEPRECATION_TEXT)
    sdk = state.sdk

    def handle_row(username):
        _remove_high_risk_employee(sdk, username)

    run_bulk_process(
        handle_row,
        csv_rows,
        progress_label="Removing users from high risk employee detection list:",
    )


@bulk.command(
    name="add-risk-tags",
    help=f"{DEPRECATION_TEXT}\n\nAdds risk tags to users in bulk using a CSV file with format: "
    f"{','.join(RISK_TAG_CSV_HEADERS)}.",
)
@read_csv_arg(headers=RISK_TAG_CSV_HEADERS)
@sdk_options()
def bulk_add_risk_tags(state, csv_rows):
    deprecation_warning(DEPRECATION_TEXT)
    sdk = state.sdk

    def handle_row(username, tag):
        _add_risk_tags(sdk, username, tag)

    run_bulk_process(
        handle_row,
        csv_rows,
        progress_label="Adding risk tags to users:",
    )


@bulk.command(
    name="remove-risk-tags",
    help=f"{DEPRECATION_TEXT}\n\nRemoves risk tags from users in bulk using a CSV file with "
    f"format: {','.join(RISK_TAG_CSV_HEADERS)}.",
)
@read_csv_arg(headers=RISK_TAG_CSV_HEADERS)
@sdk_options()
def bulk_remove_risk_tags(state, csv_rows):
    deprecation_warning(DEPRECATION_TEXT)
    sdk = state.sdk

    def handle_row(username, tag):
        _remove_risk_tags(sdk, username, tag)

    run_bulk_process(
        handle_row,
        csv_rows,
        progress_label="Removing risk tags from users:",
    )


def _get_high_risk_employees(sdk, filter):
    return sdk.detectionlists.high_risk_employee.get_all(filter)


def _add_high_risk_employee(sdk, username, cloud_alias, risk_tag, notes):
    risk_tag = handle_list_args(risk_tag)
    user_id = get_user_id(sdk, username)
    sdk.detectionlists.high_risk_employee.add(user_id)
    update_user(sdk, username, cloud_alias=cloud_alias, risk_tag=risk_tag, notes=notes)


def _remove_high_risk_employee(sdk, username):
    user_id = get_user_id(sdk, username)
    sdk.detectionlists.high_risk_employee.remove(user_id)



from code42cli.click_ext.groups import ExtensionGroup
from code42cli.main import CONTEXT_SETTINGS
from code42cli.options import debug_option
from code42cli.options import pass_state
from code42cli.options import profile_option


def sdk_options(f):
    """Decorator that adds two `click.option`s (--profile, --debug) to wrapped command, as well as
    passing the `code42cli.options.CLIState` object using the [click.make_pass_decorator](https://click.palletsprojects.com/en/7.x/api/#click.make_pass_decorator),
    which automatically instantiates the `py42` sdk using the Code42 profile provided from the `--profile`
    option. The `py42` sdk can be accessed from the `state.sdk` attribute.

    Example:

        @click.command()
        @sdk_options
        def get_current_user_command(state):
            my_user = state.sdk.users.get_current()
            print(my_user)
    """
    f = profile_option()(f)
    f = debug_option()(f)
    f = pass_state(f)
    return f


script = ExtensionGroup(context_settings=CONTEXT_SETTINGS)
"""A `click.Group` subclass that enables the Code42 CLI's custom error handling/logging to be used
in extension scripts. If only a single command is added to the `script` group it also uses that
command as the default, so the command name doesn't need to be called explicitly.

Example:

    @click.command()
    @click.argument("guid")
    @sdk_options
    def get_device_info(state, guid)
        device = state.sdk.devices.get_by_guid(guid)
        print(device)

    if __name__ == "__main__":
        script.add_command(my_command)
        script()

The script can then be invoked directly without needing to call the `get-device-info` subcommand:

    python script.py --profile my_profile <guid>
"""



from py42.__version__ import __version__ as py42version

from code42cli.__version__ import __version__ as cliversion


PRODUCT_NAME = "code42cli"
MAIN_COMMAND = "code42"
BANNER = f"""\b
 dP""b8  dP"Yb  8888b. 888888  dP88  oP"Yb.
dP   `" dP   Yb 8I  Yb 88__   dP 88  "' dP'
Yb      Yb   dP 8I  dY 88""  d888888   dP'
 YboodP  YbodP  8888Y" 888888    88  .d8888

code42cli version {cliversion}, by Code42 Software.
powered by py42 version {py42version}."""



from datetime import date

import click
import numpy as np
from pandas import concat
from pandas import DataFrame
from pandas import json_normalize
from pandas import Series
from pandas import to_datetime
from py42 import exceptions
from py42.clients.settings.device_settings import IncydrDeviceSettings
from py42.exceptions import Py42NotFoundError

from code42cli.bulk import generate_template_cmd_factory
from code42cli.bulk import run_bulk_process
from code42cli.click_ext.groups import OrderedGroup
from code42cli.click_ext.options import incompatible_with
from code42cli.click_ext.types import MagicDate
from code42cli.date_helper import round_datetime_to_day_end
from code42cli.date_helper import round_datetime_to_day_start
from code42cli.errors import Code42CLIError
from code42cli.file_readers import read_csv_arg
from code42cli.options import format_option
from code42cli.options import sdk_options
from code42cli.output_formats import DataFrameOutputFormatter
from code42cli.output_formats import OutputFormat
from code42cli.output_formats import OutputFormatter
from code42cli.worker import create_worker_stats


@click.group(cls=OrderedGroup)
@sdk_options(hidden=True)
def devices(state):
    """Manage devices within your Code42 environment."""
    pass


device_guid_argument = click.argument(
    "device-guid",
    type=str,
    callback=lambda ctx, param, arg: _verify_guid_type(arg),
)

new_device_name_option = click.option(
    "-n", "--new-device-name", help="The new name for the device.", required=True
)


def change_device_name_option(help_msg):
    return click.option(
        "--change-device-name",
        required=False,
        is_flag=True,
        default=False,
        help=help_msg,
    )


DATE_FORMAT = "%Y-%m-%d"
purge_date_option = click.option(
    "--purge-date",
    required=False,
    type=click.DateTime(formats=[DATE_FORMAT]),
    default=None,
    help="The date on which the archive should be purged from cold storage in yyyy-MM-dd format. "
    "If not provided, the date will be set according to the appropriate organization settings.",
)


@devices.command()
@device_guid_argument
@new_device_name_option
@sdk_options()
def rename(state, device_guid, new_device_name):
    """Rename a device with Code42. Requires the device GUID to rename."""
    _change_device_name(state.sdk, device_guid, new_device_name)


@devices.command()
@device_guid_argument
@change_device_name_option(
    "Prepend 'deactivated_<current_date>' to the name of the device if deactivation is successful."
)
@purge_date_option
@sdk_options()
def deactivate(state, device_guid, change_device_name, purge_date):
    """Deactivate a device within Code42. Requires the device GUID to deactivate."""
    _deactivate_device(state.sdk, device_guid, change_device_name, purge_date)


@devices.command()
@device_guid_argument
@sdk_options()
def reactivate(state, device_guid):
    """Reactivate a device within Code42. Requires the device GUID to reactivate."""
    _reactivate_device(state.sdk, device_guid)


def _deactivate_device(sdk, device_guid, change_device_name, purge_date):
    try:
        device = _change_device_activation(sdk, device_guid, "deactivate")
    except exceptions.Py42BadRequestError:
        raise Code42CLIError(f"The device with GUID '{device_guid}' is in legal hold.")
    if purge_date:
        _update_cold_storage_purge_date(sdk, device_guid, purge_date)
    if change_device_name and not device.data["name"].startswith("deactivated_"):
        _change_device_name(
            sdk,
            device_guid,
            "deactivated_"
            + date.today().strftime("%Y-%m-%d")
            + "_"
            + device.data["name"],
        )


def _reactivate_device(sdk, device_guid):
    _change_device_activation(sdk, device_guid, "reactivate")


def _change_device_activation(sdk, device_guid, cmd_str):
    try:
        device = sdk.devices.get_by_guid(device_guid)
        device_id = device.data["computerId"]
        if cmd_str == "reactivate":
            sdk.devices.reactivate(device_id)
        elif cmd_str == "deactivate":
            sdk.devices.deactivate(device_id)
        return device
    except exceptions.Py42NotFoundError:
        raise Code42CLIError(f"The device with GUID '{device_guid}' was not found.")
    except exceptions.Py42ForbiddenError:
        raise Code42CLIError(
            f"Unable to {cmd_str} the device with GUID '{device_guid}'."
        )


def _verify_guid_type(device_guid):
    if device_guid is None:
        return
    try:
        int(device_guid)
        return device_guid
    except ValueError:
        raise Code42CLIError("Not a valid GUID.")


def _update_cold_storage_purge_date(sdk, guid, purge_date):
    archives_response = sdk.archive.get_all_by_device_guid(guid)
    archive_guid_list = [
        archive["archiveGuid"]
        for page in archives_response
        for archive in page["archives"]
        if archive["format"] != "ARCHIVE_V2"
    ]
    for archive_guid in archive_guid_list:
        sdk.archive.update_cold_storage_purge_date(
            archive_guid, purge_date.strftime("%Y-%m-%d")
        )


def _change_device_name(sdk, guid, name):
    try:
        device_settings = sdk.devices.get_settings(guid)
        if isinstance(device_settings, IncydrDeviceSettings):
            raise Code42CLIError(
                "Failed to rename device. Incydr devices cannot be renamed."
            )
        device_settings.name = name
        sdk.devices.update_settings(device_settings)
    except KeyError:
        raise Code42CLIError(
            "Failed to rename device. This device is missing expected settings fields."
        )
    except exceptions.Py42ForbiddenError:
        raise Code42CLIError(
            f"You don't have the necessary permissions to rename the device with GUID '{guid}'."
        )
    except exceptions.Py42NotFoundError:
        raise Code42CLIError(f"The device with GUID '{guid}' was not found.")


@devices.command()
@device_guid_argument
@sdk_options()
def show(state, device_guid):
    """Print individual device details. Requires device GUID."""

    formatter = OutputFormatter(OutputFormat.TABLE, _device_info_keys_map())
    backup_set_formatter = OutputFormatter(OutputFormat.TABLE, _backup_set_keys_map())
    device_info = _get_device_info(state.sdk, device_guid)
    formatter.echo_formatted_list([device_info])
    backup_usage = device_info.get("backupUsage")
    if backup_usage:
        click.echo()
        backup_set_formatter.echo_formatted_list(backup_usage)


def _device_info_keys_map():
    return {
        "name": "Name",
        "osHostname": "Hostname",
        "guid": "GUID",
        "status": "Status",
        "lastConnected": "Last Connected Date",
        "productVersion": "Code42 Version",
        "osName": "Operating System",
        "osVersion": "Operating System Version",
    }


def _backup_set_keys_map():
    return {
        "targetComputerName": "Destination",
        "lastBackup": "Last Backup Activity",
        "lastCompleted": "Last Completed Backup",
        "archiveBytes": "Archive Size in Bytes",
        "archiveGuid": "Archive GUID",
    }


def _get_device_info(sdk, device_guid):
    return sdk.devices.get_by_guid(device_guid, include_backup_usage=True).data


active_option = click.option(
    "--active",
    is_flag=True,
    help="Limits results to only active devices.",
    default=None,
)
inactive_option = click.option(
    "--inactive",
    is_flag=True,
    help="Limits results to only deactivated devices.",
    cls=incompatible_with("active"),
)
org_uid_option = click.option(
    "--org-uid",
    required=False,
    type=str,
    default=None,
    help="Limit devices to only those in the organization you specify. "
    "Note that child organizations will be included.",
)

include_usernames_option = click.option(
    "--include-usernames",
    required=False,
    type=bool,
    default=False,
    is_flag=True,
    help="Add the username associated with a device to the output.",
)


@devices.command(name="list")
@active_option
@inactive_option
@org_uid_option
@click.option(
    "--include-backup-usage",
    required=False,
    type=bool,
    default=False,
    is_flag=True,
    help="Return backup usage information for each device (may significantly lengthen the size "
    "of the return).",
)
@include_usernames_option
@click.option(
    "--include-settings",
    required=False,
    type=bool,
    default=False,
    is_flag=True,
    help="Include device settings in output.",
)
@click.option(
    "--include-legal-hold-membership",
    required=False,
    type=bool,
    default=False,
    is_flag=True,
    help="Include legal hold membership in output.",
)
@click.option(
    "--include-total-storage",
    required=False,
    type=bool,
    default=False,
    is_flag=True,
    help="Include backup archive count and total storage in output.",
)
@click.option(
    "--exclude-most-recently-connected",
    type=int,
    help="Filter out the N most recently connected devices per user. "
    "Useful for identifying duplicate and/or replaced devices that are no longer needed across "
    "an environment. If a user has 2 devices and N=1, the one device with the most recent "
    "'lastConnected' date will not show up in the result list.",
)
@click.option(
    "--last-connected-before",
    type=MagicDate(rounding_func=round_datetime_to_day_start),
    help=f"Include devices only when the 'lastConnected' field is after the provided value. {MagicDate.HELP_TEXT}",
)
@click.option(
    "--last-connected-after",
    type=MagicDate(rounding_func=round_datetime_to_day_end),
    help="Include devices only when 'lastConnected' field is after the provided value. "
    "Argument format options are the same as --last-connected-before.",
)
@click.option(
    "--created-before",
    type=MagicDate(rounding_func=round_datetime_to_day_start),
    help="Include devices only when 'creationDate' field is less than the provided value. "
    "Argument format options are the same as --last-connected-before.",
)
@click.option(
    "--created-after",
    type=MagicDate(rounding_func=round_datetime_to_day_end),
    help="Include devices only when 'creationDate' field is greater than the provided value. "
    "Argument format options are the same as --last-connected-before.",
)
@format_option
@sdk_options()
def list_devices(
    state,
    active,
    inactive,
    org_uid,
    include_backup_usage,
    include_usernames,
    include_settings,
    include_legal_hold_membership,
    include_total_storage,
    exclude_most_recently_connected,
    last_connected_after,
    last_connected_before,
    created_after,
    created_before,
    format,
):
    """Get information about many devices."""
    if inactive:
        active = False
    columns = [
        "computerId",
        "guid",
        "name",
        "osHostname",
        "status",
        "lastConnected",
        "creationDate",
        "productVersion",
        "osName",
        "osVersion",
        "userUid",
    ]
    df = _get_device_dataframe(
        state.sdk,
        columns,
        active,
        org_uid,
        (include_backup_usage or include_total_storage),
    )
    if exclude_most_recently_connected:
        most_recent = (
            df.sort_values(["userUid", "lastConnected"], ascending=False)
            .groupby("userUid")
            .head(exclude_most_recently_connected)
        )
        df = df.drop(most_recent.index)
    if last_connected_after:
        df = df.loc[to_datetime(df.lastConnected) > last_connected_after]
    if last_connected_before:
        df = df.loc[to_datetime(df.lastConnected) < last_connected_before]
    if created_after:
        df = df.loc[to_datetime(df.creationDate) > created_after]
    if created_before:
        df = df.loc[to_datetime(df.creationDate) < created_before]
    if include_total_storage:
        df = _add_storage_totals_to_dataframe(df, include_backup_usage)
    if include_settings:
        df = _add_settings_to_dataframe(state.sdk, df)
    if include_usernames:
        df = _add_usernames_to_device_dataframe(state.sdk, df)
    if include_legal_hold_membership:
        df = _add_legal_hold_membership_to_device_dataframe(state.sdk, df)
    formatter = DataFrameOutputFormatter(format)
    formatter.echo_formatted_dataframes(df)


def _add_legal_hold_membership_to_device_dataframe(sdk, df):
    columns = ["legalHold.legalHoldUid", "legalHold.name", "user.userUid"]

    legal_hold_member_dataframe = (
        json_normalize(list(_get_all_active_hold_memberships(sdk)))[columns]
        .groupby(["user.userUid"])
        .agg(",".join)
        .rename(
            {
                "legalHold.legalHoldUid": "legalHoldUid",
                "legalHold.name": "legalHoldName",
            },
            axis=1,
        )
    )
    df = df.merge(
        legal_hold_member_dataframe,
        how="left",
        left_on="userUid",
        right_on="user.userUid",
    )

    df.loc[df["status"] == "Deactivated", ["legalHoldUid", "legalHoldName"]] = np.nan

    return df


def _get_all_active_hold_memberships(sdk):
    for page in sdk.legalhold.get_all_matters(active=True):
        for matter in page["legalHolds"]:
            for _page in sdk.legalhold.get_all_matter_custodians(
                legal_hold_uid=matter["legalHoldUid"], active=True
            ):
                yield from _page["legalHoldMemberships"]


def _get_device_dataframe(
    sdk, columns, active=None, org_uid=None, include_backup_usage=False
):
    devices_generator = sdk.devices.get_all(
        active=active, include_backup_usage=include_backup_usage, org_uid=org_uid
    )
    devices_list = []
    if include_backup_usage:
        columns.append("backupUsage")
    for page in devices_generator:
        devices_list.extend(page["computers"])
    return DataFrame.from_records(devices_list, columns=columns)


def _add_settings_to_dataframe(sdk, device_dataframe):
    macos_guids = [
        {"guid": value}
        for value in device_dataframe.loc[
            device_dataframe["osName"] == "mac", "guid"
        ].values
    ]

    def handle_row(guid):
        try:
            full_disk_access_status = sdk.devices.get_agent_full_disk_access_state(
                guid
            ).data[
                "value"
            ]  # returns 404 error if device isn't a Mac or doesn't have full disk access
        except Py42NotFoundError:
            full_disk_access_status = False
        return {
            "guid": guid,
            "full disk access status": full_disk_access_status,
        }

    result_list = DataFrame.from_records(
        run_bulk_process(
            handle_row, macos_guids, progress_label="Getting device settings"
        )
    )
    try:
        return device_dataframe.merge(result_list, how="left", on="guid")
    except KeyError:
        return device_dataframe


def _add_usernames_to_device_dataframe(sdk, device_dataframe):
    users_generator = sdk.users.get_all()
    users_list = []
    for page in users_generator:
        users_list.extend(page["users"])
    users_dataframe = DataFrame.from_records(
        users_list, columns=["username", "userUid"]
    )
    return device_dataframe.merge(users_dataframe, how="left", on="userUid")


def _add_storage_totals_to_dataframe(df, include_backup_usage):
    df[["archiveCount", "totalStorageBytes"]] = df["backupUsage"].apply(
        _break_backup_usage_into_total_storage
    )

    if not include_backup_usage:
        df = df.drop("backupUsage", axis=1)
    return df


def _break_backup_usage_into_total_storage(backup_usage):
    total_storage = 0
    archive_count = 0
    for archive in backup_usage:
        if archive["archiveFormat"] != "ARCHIVE_V2":
            archive_count += 1
            total_storage += archive["archiveBytes"]
    return Series([archive_count, total_storage])


@devices.command()
@active_option
@inactive_option
@org_uid_option
@include_usernames_option
@format_option
@sdk_options()
def list_backup_sets(
    state,
    active,
    inactive,
    org_uid,
    include_usernames,
    format,
):
    """Get information about many devices and their backup sets."""
    if inactive:
        active = False
    columns = ["guid", "userUid"]
    df = _get_device_dataframe(state.sdk, columns, active, org_uid)
    if include_usernames:
        df = _add_usernames_to_device_dataframe(state.sdk, df)
    df = _add_backup_set_settings_to_dataframe(state.sdk, df)
    formatter = DataFrameOutputFormatter(format)
    formatter.echo_formatted_dataframes(df)


def _add_backup_set_settings_to_dataframe(sdk, devices_dataframe):
    rows = [{"guid": guid} for guid in devices_dataframe["guid"].values]

    def handle_row(guid):
        try:
            current_device_settings = sdk.devices.get_settings(guid)
        except Exception as err:
            return DataFrame.from_records(
                [
                    {
                        "guid": guid,
                        "ERROR": f"Unable to retrieve device settings for {guid}: {err}",
                    }
                ]
            )
        current_result_dataframe = DataFrame.from_records(
            [
                {
                    "guid": current_device_settings.guid,
                    "backup set name": backup_set["name"],
                    "destinations": [
                        destination for destination in backup_set.destinations.values()
                    ],
                    "included files": list(backup_set.included_files),
                    "excluded files": list(backup_set.excluded_files),
                    "filename exclusions": list(backup_set.filename_exclusions),
                    "locked": backup_set.locked,
                }
                for backup_set in current_device_settings.backup_sets
            ]
        )
        return current_result_dataframe

    result_list = run_bulk_process(
        handle_row, rows, progress_label="Getting device settings"
    )
    try:
        return devices_dataframe.merge(concat(result_list), how="left", on="guid")
    except KeyError:
        return devices_dataframe


@devices.group(cls=OrderedGroup)
@sdk_options(hidden=True)
def bulk(state):
    """Tools for managing devices in bulk."""
    pass


_bulk_device_activation_headers = ["guid"]
_bulk_device_rename_headers = ["guid", "name"]

devices_generate_template = generate_template_cmd_factory(
    group_name="devices",
    commands_dict={
        "reactivate": _bulk_device_activation_headers,
        "deactivate": _bulk_device_activation_headers,
        "rename": _bulk_device_rename_headers,
    },
    help_message="Generate the CSV template needed for bulk device commands.",
)
bulk.add_command(devices_generate_template)


@bulk.command(name="deactivate")
@read_csv_arg(headers=_bulk_device_activation_headers)
@change_device_name_option(
    "Prepend 'deactivated_<current_date>' to the name of any successfully deactivated devices."
)
@purge_date_option
@format_option
@sdk_options()
def bulk_deactivate(state, csv_rows, change_device_name, purge_date, format):
    """Deactivate all devices from the provided CSV containing a 'guid' column."""

    # Initialize the SDK before starting any bulk processes
    # to prevent multiple instances and having to enter 2fa multiple times.
    sdk = state.sdk

    csv_rows[0]["deactivated"] = "False"
    formatter = OutputFormatter(format, {key: key for key in csv_rows[0].keys()})
    stats = create_worker_stats(len(csv_rows))
    for row in csv_rows:
        row["change_device_name"] = change_device_name
        row["purge_date"] = purge_date

    def handle_row(**row):
        try:
            _deactivate_device(
                sdk, row["guid"], row["change_device_name"], row["purge_date"]
            )
            row["deactivated"] = "True"
        except Exception as err:
            row["deactivated"] = f"False: {err}"
            stats.increment_total_errors()
        return row

    result_rows = run_bulk_process(
        handle_row,
        csv_rows,
        progress_label="Deactivating devices:",
        stats=stats,
        raise_global_error=False,
    )
    formatter.echo_formatted_list(result_rows)


@bulk.command(name="reactivate")
@read_csv_arg(headers=_bulk_device_activation_headers)
@format_option
@sdk_options()
def bulk_reactivate(state, csv_rows, format):
    """Reactivate all devices from the provided CSV containing a 'guid' column."""

    # Initialize the SDK before starting any bulk processes
    # to prevent multiple instances and having to enter 2fa multiple times.
    sdk = state.sdk

    csv_rows[0]["reactivated"] = "False"
    formatter = OutputFormatter(format, {key: key for key in csv_rows[0].keys()})
    stats = create_worker_stats(len(csv_rows))

    def handle_row(**row):
        try:
            _reactivate_device(sdk, row["guid"])
            row["reactivated"] = "True"
        except Exception as err:
            row["reactivated"] = f"False: {err}"
            stats.increment_total_errors()
        return row

    result_rows = run_bulk_process(
        handle_row,
        csv_rows,
        progress_label="Reactivating devices:",
        stats=stats,
        raise_global_error=False,
    )
    formatter.echo_formatted_list(result_rows)


@bulk.command(name="rename")
@read_csv_arg(headers=_bulk_device_rename_headers)
@format_option
@sdk_options()
def bulk_rename(state, csv_rows, format):
    """Rename all devices from the provided CSV containing a 'guid' and a 'name' column."""

    # Initialize the SDK before starting any bulk processes
    # to prevent multiple instances and having to enter 2fa multiple times.
    sdk = state.sdk

    csv_rows[0]["renamed"] = "False"
    formatter = OutputFormatter(format, {key: key for key in csv_rows[0].keys()})
    stats = create_worker_stats(len(csv_rows))

    def handle_row(**row):
        try:
            _change_device_name(sdk, row["guid"], row["name"])
            row["renamed"] = "True"
        except Exception as err:
            row["renamed"] = f"False: {err}"
            stats.increment_total_errors()
        return row

    result_rows = run_bulk_process(
        handle_row,
        csv_rows,
        progress_label="Renaming devices:",
        stats=stats,
        raise_global_error=False,
    )
    formatter.echo_formatted_list(result_rows)



import click

from code42cli.click_ext.types import MagicDate
from code42cli.click_ext.types import TOTP
from code42cli.cmds.search.options import AdvancedQueryAndSavedSearchIncompatible
from code42cli.cmds.search.options import BeginOption
from code42cli.date_helper import convert_datetime_to_timestamp
from code42cli.date_helper import round_datetime_to_day_end
from code42cli.date_helper import round_datetime_to_day_start
from code42cli.enums import OutputFormat
from code42cli.enums import SendToFileEventsOutputFormat
from code42cli.errors import Code42CLIError
from code42cli.logger.enums import ServerProtocol
from code42cli.profile import get_profile
from code42cli.sdk_client import create_sdk


def yes_option(hidden=False):
    return click.option(
        "-y",
        "--assume-yes",
        is_flag=True,
        expose_value=False,
        callback=lambda ctx, param, value: ctx.obj.set_assume_yes(value),
        help='Assume "yes" as the answer to all prompts and run non-interactively.',
        hidden=hidden,
    )


format_option = click.option(
    "-f",
    "--format",
    type=click.Choice(OutputFormat(), case_sensitive=False),
    help="The output format of the result. Defaults to table format.",
    default=OutputFormat.TABLE,
)


class CLIState:
    def __init__(self):
        try:
            self._profile = get_profile()
        except Code42CLIError:
            self._profile = None
        self.totp = None
        self.debug = False
        self._sdk = None
        self.search_filters = []
        self.assume_yes = False

    @property
    def profile(self):
        if self._profile is None:
            self._profile = get_profile()
        return self._profile

    @profile.setter
    def profile(self, value):
        self._profile = value

    @property
    def sdk(self):
        if self._sdk is None:
            self._sdk = create_sdk(self.profile, self.debug, totp=self.totp)
        return self._sdk

    def set_assume_yes(self, param):
        self.assume_yes = param


def set_profile(ctx, param, value):
    """Sets the profile on the global state object when --profile <name> is passed to commands
    decorated with @sdk_options."""
    if value:
        ctx.ensure_object(CLIState).profile = get_profile(value)


def set_debug(ctx, param, value):
    """Sets debug to True on global state object when --debug/-d is passed to commands decorated
    with @sdk_options.
    """
    if value:
        ctx.ensure_object(CLIState).debug = value


def set_totp(ctx, param, value):
    """Sets TOTP token on global state object for multi-factor authentication."""
    if value:
        ctx.ensure_object(CLIState).totp = value


def profile_option(hidden=False):
    opt = click.option(
        "--profile",
        expose_value=False,
        callback=set_profile,
        hidden=hidden,
        help="The name of the Code42 CLI profile to use when executing this command.",
    )
    return opt


def debug_option(hidden=False):
    opt = click.option(
        "-d",
        "--debug",
        is_flag=True,
        expose_value=False,
        callback=set_debug,
        hidden=hidden,
        help="Turn on debug logging.",
    )
    return opt


def totp_option(hidden=False):
    opt = click.option(
        "--totp",
        type=TOTP(),
        expose_value=False,
        callback=set_totp,
        hidden=hidden,
        help="TOTP token for multi-factor authentication.",
    )
    return opt


pass_state = click.make_pass_decorator(CLIState, ensure=True)


def sdk_options(hidden=False):
    def decorator(f):
        f = profile_option(hidden)(f)
        f = totp_option(hidden)(f)
        f = debug_option(hidden)(f)
        f = pass_state(f)
        return f

    return decorator


def server_options(f):
    hostname_arg = click.argument("hostname")
    protocol_option = click.option(
        "-p",
        "--protocol",
        type=click.Choice(ServerProtocol(), case_sensitive=False),
        default=ServerProtocol.UDP,
        help="Protocol used to send logs to server. Defaults to UDP.",
    )
    f = hostname_arg(f)
    f = protocol_option(f)
    return f


send_to_format_options = click.option(
    "-f",
    "--format",
    type=click.Choice(SendToFileEventsOutputFormat(), case_sensitive=False),
    help="The output format of the result. Defaults to json format.",
    default=SendToFileEventsOutputFormat.RAW,
)


def begin_option(term, **kwargs):
    defaults = dict(
        type=MagicDate(rounding_func=round_datetime_to_day_start),
        help=f"The beginning of the date range in which to look for {term}. {MagicDate.HELP_TEXT} [required unless --use-checkpoint option used]",
        cls=BeginOption,
        callback=lambda ctx, param, arg: convert_datetime_to_timestamp(arg),
    )
    defaults.update(kwargs)
    return click.option("-b", "--begin", **defaults)


def end_option(term, **kwargs):
    defaults = dict(
        type=MagicDate(rounding_func=round_datetime_to_day_end),
        cls=AdvancedQueryAndSavedSearchIncompatible,
        help=f"The end of the date range in which to look for {term}, argument format options are "
        "the same as `--begin`.",
        callback=lambda ctx, param, arg: convert_datetime_to_timestamp(arg),
    )
    defaults.update(kwargs)
    return click.option("-e", "--end", **defaults)


def checkpoint_option(term, **kwargs):
    defaults = dict(
        help=f"Use a checkpoint with the given name to only get {term} that were not previously retrieved."
        f"If a checkpoint for {term} with the given name doesn't exist, it will be created on the first run."
        "Subsequent CLI runs with this flag and the same name will use the stored checkpoint to modify the search query and then update the stored checkpoint"
    )
    defaults.update(kwargs)
    return click.option("-c", "--use-checkpoint", **defaults)


def set_begin_default_dict(term):
    return dict(
        type=MagicDate(rounding_func=round_datetime_to_day_start),
        help=f"The beginning of the date range in which to look for {term}. {MagicDate.HELP_TEXT}",
        callback=lambda ctx, param, arg: convert_datetime_to_timestamp(arg),
    )


def set_end_default_dict(term):
    return dict(
        type=MagicDate(rounding_func=round_datetime_to_day_end),
        help=f"The end of the date range in which to look for {term}, argument format options are "
        "the same as `--begin`.",
        callback=lambda ctx, param, arg: convert_datetime_to_timestamp(arg),
    )


column_option = click.option(
    "--columns",
    default=None,
    callback=lambda ctx, param, value: value.split(",") if value is not None else None,
    help="Filter output to include only specified columns. Accepts comma-separated list of column names (case-insensitive).",
)



import datetime
from pprint import pformat

import click
import py42.sdk.queries.fileevents.filters as f
from click import echo
from pandas import DataFrame
from py42.exceptions import Py42InvalidPageTokenError
from py42.sdk.queries.fileevents.file_event_query import FileEventQuery
from py42.sdk.queries.fileevents.filters import InsertionTimestamp
from py42.sdk.queries.fileevents.filters.exposure_filter import ExposureType
from py42.sdk.queries.fileevents.filters.file_filter import FileCategory
from py42.sdk.queries.fileevents.filters.risk_filter import RiskIndicator
from py42.sdk.queries.fileevents.filters.risk_filter import RiskSeverity

import code42cli.cmds.search.options as searchopt
import code42cli.options as opt
from code42cli.click_ext.groups import OrderedGroup
from code42cli.click_ext.options import incompatible_with
from code42cli.click_ext.types import MapChoice
from code42cli.cmds.search import SendToCommand
from code42cli.cmds.search.cursor_store import FileEventCursorStore
from code42cli.cmds.util import convert_to_or_query
from code42cli.cmds.util import create_time_range_filter
from code42cli.date_helper import convert_datetime_to_timestamp
from code42cli.date_helper import limit_date_range
from code42cli.enums import OutputFormat
from code42cli.logger import get_main_cli_logger
from code42cli.options import column_option
from code42cli.options import format_option
from code42cli.options import sdk_options
from code42cli.output_formats import DataFrameOutputFormatter
from code42cli.output_formats import FileEventsOutputFormat
from code42cli.output_formats import FileEventsOutputFormatter
from code42cli.util import warn_interrupt

logger = get_main_cli_logger()
MAX_EVENT_PAGE_SIZE = 10000

SECURITY_DATA_KEYWORD = "file events"
file_events_format_option = click.option(
    "-f",
    "--format",
    type=click.Choice(FileEventsOutputFormat(), case_sensitive=False),
    help="The output format of the result. Defaults to table format.",
    default=FileEventsOutputFormat.TABLE,
)
exposure_type_option = click.option(
    "-t",
    "--type",
    multiple=True,
    type=click.Choice(list(ExposureType.choices())),
    cls=searchopt.AdvancedQueryAndSavedSearchIncompatible,
    callback=searchopt.is_in_filter(f.ExposureType),
    help="Limits events to those with given exposure types.",
)
username_option = click.option(
    "--c42-username",
    multiple=True,
    callback=searchopt.is_in_filter(f.DeviceUsername),
    cls=searchopt.AdvancedQueryAndSavedSearchIncompatible,
    help="Limits events to endpoint events for these Code42 users.",
)
actor_option = click.option(
    "--actor",
    multiple=True,
    callback=searchopt.is_in_filter(f.Actor),
    cls=searchopt.AdvancedQueryAndSavedSearchIncompatible,
    help="Limits events to only those enacted by the cloud service user "
    "of the person who caused the event.",
)
md5_option = click.option(
    "--md5",
    multiple=True,
    callback=searchopt.is_in_filter(f.MD5),
    cls=searchopt.AdvancedQueryAndSavedSearchIncompatible,
    help="Limits events to file events where the file has one of these MD5 hashes.",
)
sha256_option = click.option(
    "--sha256",
    multiple=True,
    callback=searchopt.is_in_filter(f.SHA256),
    cls=searchopt.AdvancedQueryAndSavedSearchIncompatible,
    help="Limits events to file events where the file has one of these SHA256 hashes.",
)
source_option = click.option(
    "--source",
    multiple=True,
    callback=searchopt.is_in_filter(f.Source),
    cls=searchopt.AdvancedQueryAndSavedSearchIncompatible,
    help="Limits events to only those from one of these sources. For example, Gmail, Box, or Endpoint.",
)
file_name_option = click.option(
    "--file-name",
    multiple=True,
    callback=searchopt.is_in_filter(f.FileName),
    cls=searchopt.AdvancedQueryAndSavedSearchIncompatible,
    help="Limits events to file events where the file has one of these names.",
)
file_path_option = click.option(
    "--file-path",
    multiple=True,
    callback=searchopt.is_in_filter(f.FilePath),
    cls=searchopt.AdvancedQueryAndSavedSearchIncompatible,
    help="Limits events to file events where the file is located at one of these paths. Applies to endpoint file events only.",
)
file_category_option = click.option(
    "--file-category",
    multiple=True,
    type=MapChoice(
        choices=list(FileCategory.choices()),
        extras_map={
            "AUDIO": FileCategory.AUDIO,
            "DOCUMENT": FileCategory.DOCUMENT,
            "EXECUTABLE": FileCategory.EXECUTABLE,
            "IMAGE": FileCategory.IMAGE,
            "PDF": FileCategory.PDF,
            "PRESENTATION": FileCategory.PRESENTATION,
            "SCRIPT": FileCategory.SCRIPT,
            "SOURCE_CODE": FileCategory.SOURCE_CODE,
            "SPREADSHEET": FileCategory.SPREADSHEET,
            "VIDEO": FileCategory.VIDEO,
            "VIRTUAL_DISK_IMAGE": FileCategory.VIRTUAL_DISK_IMAGE,
            "ARCHIVE": FileCategory.ZIP,
            "ZIP": FileCategory.ZIP,
            "Zip": FileCategory.ZIP,
        },
    ),
    callback=searchopt.is_in_filter(f.FileCategory),
    cls=searchopt.AdvancedQueryAndSavedSearchIncompatible,
    help="Limits events to file events where the file can be classified by one of these categories.",
)
process_owner_option = click.option(
    "--process-owner",
    multiple=True,
    callback=searchopt.is_in_filter(f.ProcessOwner),
    cls=searchopt.AdvancedQueryAndSavedSearchIncompatible,
    help="Limits exposure events by process owner, as reported by the device’s operating system. "
    "Applies only to `Printed` and `Browser or app read` events.",
)
tab_url_option = click.option(
    "--tab-url",
    multiple=True,
    callback=searchopt.is_in_filter(f.TabURL),
    cls=searchopt.AdvancedQueryAndSavedSearchIncompatible,
    help="Limits events to be exposure events with one of the specified destination tab URLs.",
)
include_non_exposure_option = click.option(
    "--include-non-exposure",
    is_flag=True,
    callback=searchopt.exists_filter(f.ExposureType),
    cls=incompatible_with(["advanced_query", "type", "saved_search"]),
    help="Get all events including non-exposure events.",
)
risk_indicator_map = {
    "PUBLIC_CORPORATE_BOX": RiskIndicator.CloudDataExposures.PUBLIC_CORPORATE_BOX,
    "PUBLIC_CORPORATE_GOOGLE": RiskIndicator.CloudDataExposures.PUBLIC_CORPORATE_GOOGLE_DRIVE,
    "PUBLIC_CORPORATE_ONEDRIVE": RiskIndicator.CloudDataExposures.PUBLIC_CORPORATE_ONEDRIVE,
    "SENT_CORPORATE_GMAIL": RiskIndicator.CloudDataExposures.SENT_CORPORATE_GMAIL,
    "SHARED_CORPORATE_BOX": RiskIndicator.CloudDataExposures.SHARED_CORPORATE_BOX,
    "SHARED_CORPORATE_GOOGLE_DRIVE": RiskIndicator.CloudDataExposures.SHARED_CORPORATE_GOOGLE_DRIVE,
    "SHARED_CORPORATE_ONEDRIVE": RiskIndicator.CloudDataExposures.SHARED_CORPORATE_ONEDRIVE,
    "AMAZON_DRIVE": RiskIndicator.CloudStorageUploads.AMAZON_DRIVE,
    "BOX": RiskIndicator.CloudStorageUploads.BOX,
    "DROPBOX": RiskIndicator.CloudStorageUploads.DROPBOX,
    "GOOGLE_DRIVE": RiskIndicator.CloudStorageUploads.GOOGLE_DRIVE,
    "ICLOUD": RiskIndicator.CloudStorageUploads.ICLOUD,
    "MEGA": RiskIndicator.CloudStorageUploads.MEGA,
    "ONEDRIVE": RiskIndicator.CloudStorageUploads.ONEDRIVE,
    "ZOHO": RiskIndicator.CloudStorageUploads.ZOHO,
    "BITBUCKET": RiskIndicator.CodeRepositoryUploads.BITBUCKET,
    "GITHUB": RiskIndicator.CodeRepositoryUploads.GITHUB,
    "GITLAB": RiskIndicator.CodeRepositoryUploads.GITLAB,
    "SOURCEFORGE": RiskIndicator.CodeRepositoryUploads.SOURCEFORGE,
    "STASH": RiskIndicator.CodeRepositoryUploads.STASH,
    "163.COM": RiskIndicator.EmailServiceUploads.ONESIXTHREE_DOT_COM,
    "126.COM": RiskIndicator.EmailServiceUploads.ONETWOSIX_DOT_COM,
    "AOL": RiskIndicator.EmailServiceUploads.AOL,
    "COMCAST": RiskIndicator.EmailServiceUploads.COMCAST,
    "GMAIL": RiskIndicator.EmailServiceUploads.GMAIL,
    "ICLOUD_MAIL": RiskIndicator.EmailServiceUploads.ICLOUD,
    "MAIL.COM": RiskIndicator.EmailServiceUploads.MAIL_DOT_COM,
    "OUTLOOK": RiskIndicator.EmailServiceUploads.OUTLOOK,
    "PROTONMAIL": RiskIndicator.EmailServiceUploads.PROTONMAIL,
    "QQMAIL": RiskIndicator.EmailServiceUploads.QQMAIL,
    "SINA_MAIL": RiskIndicator.EmailServiceUploads.SINA_MAIL,
    "SOHU_MAIL": RiskIndicator.EmailServiceUploads.SOHU_MAIL,
    "YAHOO": RiskIndicator.EmailServiceUploads.YAHOO,
    "ZOHO_MAIL": RiskIndicator.EmailServiceUploads.ZOHO_MAIL,
    "AIRDROP": RiskIndicator.ExternalDevices.AIRDROP,
    "REMOVABLE_MEDIA": RiskIndicator.ExternalDevices.REMOVABLE_MEDIA,
    "AUDIO": RiskIndicator.FileCategories.AUDIO,
    "DOCUMENT": RiskIndicator.FileCategories.DOCUMENT,
    "EXECUTABLE": RiskIndicator.FileCategories.EXECUTABLE,
    "IMAGE": RiskIndicator.FileCategories.IMAGE,
    "PDF": RiskIndicator.FileCategories.PDF,
    "PRESENTATION": RiskIndicator.FileCategories.PRESENTATION,
    "SCRIPT": RiskIndicator.FileCategories.SCRIPT,
    "SOURCE_CODE": RiskIndicator.FileCategories.SOURCE_CODE,
    "SPREADSHEET": RiskIndicator.FileCategories.SPREADSHEET,
    "VIDEO": RiskIndicator.FileCategories.VIDEO,
    "VIRTUAL_DISK_IMAGE": RiskIndicator.FileCategories.VIRTUAL_DISK_IMAGE,
    "ZIP": RiskIndicator.FileCategories.ZIP,
    "FACEBOOK_MESSENGER": RiskIndicator.MessagingServiceUploads.FACEBOOK_MESSENGER,
    "MICROSOFT_TEAMS": RiskIndicator.MessagingServiceUploads.MICROSOFT_TEAMS,
    "SLACK": RiskIndicator.MessagingServiceUploads.SLACK,
    "WHATSAPP": RiskIndicator.MessagingServiceUploads.WHATSAPP,
    "OTHER": RiskIndicator.Other.OTHER,
    "UNKNOWN": RiskIndicator.Other.UNKNOWN,
    "FACEBOOK": RiskIndicator.SocialMediaUploads.FACEBOOK,
    "LINKEDIN": RiskIndicator.SocialMediaUploads.LINKEDIN,
    "REDDIT": RiskIndicator.SocialMediaUploads.REDDIT,
    "TWITTER": RiskIndicator.SocialMediaUploads.TWITTER,
    "FILE_MISMATCH": RiskIndicator.UserBehavior.FILE_MISMATCH,
    "OFF_HOURS": RiskIndicator.UserBehavior.OFF_HOURS,
    "REMOTE": RiskIndicator.UserBehavior.REMOTE,
    "FIRST_DESTINATION_USE": RiskIndicator.UserBehavior.FIRST_DESTINATION_USE,
    "RARE_DESTINATION_USE": RiskIndicator.UserBehavior.RARE_DESTINATION_USE,
}
risk_indicator_map_reversed = {v: k for k, v in risk_indicator_map.items()}


def risk_indicator_callback(filter_cls):
    def callback(ctx, param, arg):
        if arg:
            mapped_args = tuple(risk_indicator_map[i] for i in arg)
            filter_func = searchopt.is_in_filter(filter_cls)
            return filter_func(ctx, param, mapped_args)

    return callback


risk_indicator_option = click.option(
    "--risk-indicator",
    multiple=True,
    type=MapChoice(
        choices=list(risk_indicator_map.keys()),
        extras_map=risk_indicator_map_reversed,
    ),
    callback=risk_indicator_callback(f.RiskIndicator),
    cls=searchopt.AdvancedQueryAndSavedSearchIncompatible,
    help="Limits events to those classified by the given risk indicator categories.",
)
risk_severity_option = click.option(
    "--risk-severity",
    multiple=True,
    type=click.Choice(list(RiskSeverity.choices())),
    callback=searchopt.is_in_filter(f.RiskSeverity),
    cls=searchopt.AdvancedQueryAndSavedSearchIncompatible,
    help="Limits events to those classified by the given risk severity.",
)
begin_option = opt.begin_option(
    SECURITY_DATA_KEYWORD,
    callback=lambda ctx, param, arg: convert_datetime_to_timestamp(
        limit_date_range(arg, max_days_back=90)
    ),
)
end_option = opt.end_option(SECURITY_DATA_KEYWORD)
checkpoint_option = opt.checkpoint_option(
    SECURITY_DATA_KEYWORD, cls=searchopt.AdvancedQueryAndSavedSearchIncompatible
)
advanced_query_option = searchopt.advanced_query_option(SECURITY_DATA_KEYWORD)


def _get_saved_search_option():
    def _get_saved_search_query(ctx, param, arg):
        if arg is None:
            return
        query = ctx.obj.sdk.securitydata.savedsearches.get_query(arg)
        return query

    return click.option(
        "--saved-search",
        help="Get events from a saved search filter with the given ID."
        "WARNING: Using a saved search is incompatible with other query-building arguments.",
        callback=_get_saved_search_query,
        cls=incompatible_with("advanced_query"),
    )


def search_options(f):
    f = column_option(f)
    f = checkpoint_option(f)
    f = advanced_query_option(f)
    f = searchopt.or_query_option(f)
    f = end_option(f)
    f = begin_option(f)
    return f


def file_event_options(f):
    f = exposure_type_option(f)
    f = username_option(f)
    f = actor_option(f)
    f = md5_option(f)
    f = sha256_option(f)
    f = source_option(f)
    f = file_name_option(f)
    f = file_path_option(f)
    f = file_category_option(f)
    f = process_owner_option(f)
    f = tab_url_option(f)
    f = include_non_exposure_option(f)
    f = risk_indicator_option(f)
    f = risk_severity_option(f)
    f = _get_saved_search_option()(f)
    return f


@click.group(cls=OrderedGroup)
@sdk_options(hidden=True)
def security_data(state):
    """Get and send file event data."""
    # store cursor getter on the group state so shared --begin option can use it in validation
    state.cursor_getter = _get_file_event_cursor_store


@security_data.command()
@click.argument("checkpoint-name")
@sdk_options()
def clear_checkpoint(state, checkpoint_name):
    """Remove the saved file event checkpoint from `--use-checkpoint/-c` mode."""
    _get_file_event_cursor_store(state.profile.name).delete(checkpoint_name)


@security_data.command()
@file_event_options
@search_options
@sdk_options()
@column_option
@searchopt.include_all_option
@file_events_format_option
def search(
    state,
    format,
    begin,
    end,
    advanced_query,
    use_checkpoint,
    saved_search,
    or_query,
    columns,
    include_all,
    **kwargs,
):

    """Search for file events."""
    if format == FileEventsOutputFormat.CEF and columns:
        raise click.BadOptionUsage(
            "columns", "--columns option can't be used with CEF format."
        )
    # set default table columns
    if format == OutputFormat.TABLE:
        if not columns and not include_all:
            columns = [
                "fileName",
                "filePath",
                "eventType",
                "eventTimestamp",
                "fileCategory",
                "fileSize",
                "fileOwner",
                "md5Checksum",
                "sha256Checksum",
                "riskIndicators",
                "riskSeverity",
            ]

    if use_checkpoint:
        cursor = _get_file_event_cursor_store(state.profile.name)
        checkpoint = _handle_timestamp_checkpoint(cursor.get(use_checkpoint), state)

        def checkpoint_func(event):
            cursor.replace(use_checkpoint, event["eventId"])

    else:
        checkpoint = checkpoint_func = None

    query = _construct_query(state, begin, end, saved_search, advanced_query, or_query)
    dfs = _get_all_file_events(state, query, checkpoint)
    formatter = FileEventsOutputFormatter(format, checkpoint_func=checkpoint_func)
    # sending to pager when checkpointing can be inaccurate due to pager buffering, so disallow pager
    force_no_pager = use_checkpoint
    formatter.echo_formatted_dataframes(
        dfs, columns=columns, force_no_pager=force_no_pager
    )


@security_data.command(cls=SendToCommand)
@file_event_options
@search_options
@sdk_options()
@searchopt.server_options
@searchopt.send_to_format_options
def send_to(
    state,
    begin,
    end,
    advanced_query,
    use_checkpoint,
    saved_search,
    or_query,
    columns,
    **kwargs,
):
    """Send events to the given server address.

    HOSTNAME format: address:port where port is optional and defaults to 514.
    """
    if use_checkpoint:
        cursor = _get_file_event_cursor_store(state.profile.name)
        checkpoint = _handle_timestamp_checkpoint(cursor.get(use_checkpoint), state)

        def checkpoint_func(event):
            cursor.replace(use_checkpoint, event["eventId"])

    else:
        checkpoint = checkpoint_func = None

    query = _construct_query(state, begin, end, saved_search, advanced_query, or_query)
    dfs = _get_all_file_events(state, query, checkpoint)
    formatter = FileEventsOutputFormatter(None, checkpoint_func=checkpoint_func)

    with warn_interrupt():
        event = None
        for event in formatter.iter_rows(dfs, columns=columns):
            state.logger.info(event)
        if event is None:  # generator was empty
            click.echo("No results found.")


@security_data.group(cls=OrderedGroup)
@sdk_options()
def saved_search(state):
    """Search for file events using saved searches."""
    pass


@saved_search.command("list")
@format_option
@sdk_options()
def _list(state, format=None):
    """List available saved searches."""
    formatter = DataFrameOutputFormatter(format)
    response = state.sdk.securitydata.savedsearches.get()
    saved_searches_df = DataFrame(response["searches"])
    formatter.echo_formatted_dataframes(
        saved_searches_df, columns=["name", "id", "notes"]
    )


@saved_search.command()
@click.argument("search-id")
@sdk_options()
def show(state, search_id):
    """Get the details of a saved search."""
    response = state.sdk.securitydata.savedsearches.get_by_id(search_id)
    echo(pformat(response["searches"]))


def _get_file_event_cursor_store(profile_name):
    return FileEventCursorStore(profile_name)


def _construct_query(state, begin, end, saved_search, advanced_query, or_query):

    if advanced_query:
        state.search_filters = advanced_query
    elif saved_search:
        state.search_filters = saved_search._filter_group_list
    else:
        if begin or end:
            state.search_filters.append(
                create_time_range_filter(f.EventTimestamp, begin, end)
            )
    if or_query:
        state.search_filters = convert_to_or_query(state.search_filters)

    if not state.search_filters:
        # if a checkpoint and _only_ --include-non-exposure is passed, the filter list will be empty, which isn't a
        # valid query, so in that case we want to fallback to a 90 day (max event age) date range. The checkpoint will
        # still cause the query results to only contain events after the checkpointed event.
        _90_days = datetime.datetime.utcnow() - datetime.timedelta(days=90)
        timestamp = convert_datetime_to_timestamp(_90_days)
        state.search_filters.append(
            create_time_range_filter(f.EventTimestamp, timestamp, None)
        )
    query = FileEventQuery(*state.search_filters)
    query.page_size = MAX_EVENT_PAGE_SIZE
    query.sort_direction = "asc"
    query.sort_key = "insertionTimestamp"
    return query


def _get_all_file_events(state, query, checkpoint=""):
    try:
        response = state.sdk.securitydata.search_all_file_events(
            query, page_token=checkpoint
        )
    except Py42InvalidPageTokenError:
        response = state.sdk.securitydata.search_all_file_events(query)
    yield DataFrame(response["fileEvents"])
    while response["nextPgToken"]:
        response = state.sdk.securitydata.search_all_file_events(
            query, page_token=response["nextPgToken"]
        )
        yield DataFrame(response["fileEvents"])


def _handle_timestamp_checkpoint(checkpoint, state):
    try:
        checkpoint = float(checkpoint)
        state.search_filters.append(InsertionTimestamp.on_or_after(checkpoint))
        return None
    except (ValueError, TypeError):
        return checkpoint



import json
import os
from os import path

from code42cli.errors import Code42CLIError
from code42cli.util import get_user_project_path


class Cursor:
    def __init__(self, location):
        self._location = location
        self._name = path.basename(location)

    @property
    def name(self):
        return self._name

    @property
    def value(self):
        with open(self._location) as checkpoint:
            return checkpoint.read()


class BaseCursorStore:
    def __init__(self, dir_path):
        self._dir_path = dir_path

    def get(self, cursor_name):
        """Gets the last stored date observed timestamp."""
        try:
            location = path.join(self._dir_path, cursor_name)
            with open(location) as checkpoint:
                checkpoint_value = checkpoint.read()
                if not checkpoint_value:
                    return None
                try:
                    return float(checkpoint_value)
                except ValueError:
                    raise Code42CLIError(
                        f"Unable to parse checkpoint from {location}, expected a unix-epoch timestamp, got '{checkpoint_value}'."
                    )
        except FileNotFoundError:
            return None

    def replace(self, cursor_name, new_checkpoint):
        """Replaces the last stored date observed timestamp with the given one."""
        location = path.join(self._dir_path, cursor_name)
        with open(location, "w") as checkpoint:
            checkpoint.write(str(new_checkpoint))

    def delete(self, cursor_name):
        """Removes a single cursor from the store."""
        try:
            location = path.join(self._dir_path, cursor_name)
            os.remove(location)
        except FileNotFoundError:
            msg = f"No checkpoint named {cursor_name} exists for this profile."
            raise Code42CLIError(msg)

    def clean(self):
        """Removes all cursors from this store."""
        cursors = self.get_all_cursors()
        for cursor in cursors:
            self.delete(cursor.name)

    def get_all_cursors(self):
        """Returns a list of all cursors stored in this directory (which is typically scoped to a profile)."""
        dir_contents = os.listdir(self._dir_path)
        return [Cursor(f) for f in dir_contents if self._is_file(f)]

    def _is_file(self, node_name):
        return path.isfile(path.join(self._dir_path, node_name))


class FileEventCursorStore(BaseCursorStore):
    def __init__(self, profile_name):
        dir_path = get_user_project_path("file_event_checkpoints", profile_name)
        super().__init__(dir_path)

    def get(self, cursor_name):
        """Gets the last stored date observed timestamp."""
        try:
            location = path.join(self._dir_path, cursor_name)
            with open(location) as checkpoint:
                checkpoint_value = checkpoint.read()
                if not checkpoint_value:
                    return None
                return str(checkpoint_value)
        except FileNotFoundError:
            return None


class AlertCursorStore(BaseCursorStore):
    def __init__(self, profile_name):
        dir_path = get_user_project_path("alert_checkpoints", profile_name)
        super().__init__(dir_path)

    def get_alerts(self, cursor_name):
        try:
            location = path.join(self._dir_path, cursor_name) + "_alerts"
            with open(location) as checkpoint:
                try:
                    return json.loads(checkpoint.read())
                except json.JSONDecodeError:
                    return []
        except FileNotFoundError:
            return []

    def replace_alerts(self, cursor_name, new_alerts):
        location = path.join(self._dir_path, cursor_name) + "_alerts"
        with open(location, "w") as checkpoint:
            checkpoint.write(json.dumps(new_alerts))


class AuditLogCursorStore(BaseCursorStore):
    def __init__(self, profile_name):
        dir_path = get_user_project_path("audit_log_checkpoints", profile_name)
        super().__init__(dir_path)

    def get_events(self, cursor_name):
        try:
            location = path.join(self._dir_path, cursor_name) + "_events"
            with open(location) as checkpoint:
                try:
                    return json.loads(checkpoint.read())
                except json.JSONDecodeError:
                    return []
        except FileNotFoundError:
            return []

    def replace_events(self, cursor_name, new_events):
        location = path.join(self._dir_path, cursor_name) + "_events"
        with open(location, "w") as checkpoint:
            checkpoint.write(json.dumps(new_events))


def get_all_cursor_stores_for_profile(profile_name):
    return [
        FileEventCursorStore(profile_name),
        AlertCursorStore(profile_name),
        AuditLogCursorStore(profile_name),
    ]



import re
from datetime import datetime
from datetime import timedelta
from datetime import timezone

import click

TIMESTAMP_REGEX = re.compile(r"(\d{4}-\d{2}-\d{2})\s*(.*)?")
MAGIC_TIME_REGEX = re.compile(r"(\d+)([dhm])$")

_FORMAT_VALUE_ERROR_MESSAGE = (
    "input must be a date/time string (e.g. 'yyyy-MM-dd', "
    "'yy-MM-dd HH:MM', 'yy-MM-dd HH:MM:SS'), or a short value in days, "
    "hours, or minutes (e.g. 30d, 24h, 15m)"
)


def convert_datetime_to_timestamp(dt):
    if dt is None:
        return
    return dt.replace(tzinfo=timezone.utc).timestamp()


def verify_timestamp_order(
    min_timestamp, max_timestamp, min_param=("-b", "--begin"), max_param="--end"
):
    if min_timestamp is None or max_timestamp is None:
        return
    if min_timestamp >= max_timestamp:
        raise click.BadParameter(
            param_hint=min_param, message=f"cannot be after {max_param} date."
        )


def limit_date_range(dt, max_days_back=90, param=None):
    if dt is None:
        return
    now = datetime.utcnow().replace(tzinfo=timezone.utc)
    if now - dt > timedelta(days=max_days_back):
        raise click.BadParameter(
            message=f"must be within {max_days_back} days.", param=param
        )
    return dt


def round_datetime_to_day_start(dt):
    return dt.replace(hour=0, minute=0, second=0, microsecond=0)


def round_datetime_to_day_end(dt):
    return dt.replace(hour=23, minute=59, second=59, microsecond=999999)



import click
from py42.services.detectionlists import _DetectionListFilters

from code42cli.cmds.shared import get_user_id
from code42cli.output_formats import OutputFormat
from code42cli.output_formats import OutputFormatter


ALL_FILTER = "ALL"


def get_choices(filters):
    filters.remove(_DetectionListFilters.OPEN)
    filters.append(ALL_FILTER)
    return filters


def handle_filter_choice(choice):
    if choice == ALL_FILTER:
        return _DetectionListFilters.OPEN
    return choice


def list_employees(employee_generator, output_format, additional_header_items=None):
    additional_header_items = additional_header_items or {}
    header = {"userName": "Username", "notes": "Notes", **additional_header_items}
    employee_list = []
    for employees in employee_generator:
        for employee in employees["items"]:
            if employee.get("notes") and output_format == OutputFormat.TABLE:
                employee["notes"] = (
                    employee["notes"].replace("\n", "\\n").replace("\t", "\\t")
                )
            employee_list.append(employee)
    if employee_list:
        formatter = OutputFormatter(output_format, header)
        formatter.echo_formatted_list(employee_list)
    else:
        click.echo("No users found.")


def update_user(sdk, username, cloud_alias=None, risk_tag=None, notes=None):
    """Updates a detection list user.

    Args:
        sdk (py42.sdk.SDKClient): py42 sdk.
        username (str): The username of the user to update.
        cloud_alias (str): A cloud alias to add to the user.
        risk_tag (iter[str]): A list of risk tags associated with user.
        notes (str): Notes about the user.
    """
    user_id = get_user_id(sdk, username)
    _update_cloud_alias(sdk, user_id, cloud_alias)
    _update_risk_tags(sdk, username, risk_tag)
    _update_notes(sdk, user_id, notes)


def _update_cloud_alias(sdk, user_id, cloud_alias):
    if cloud_alias:
        profile = sdk.detectionlists.get_user_by_id(user_id)
        cloud_aliases = profile.data.get("cloudUsernames") or []
        for alias in cloud_aliases:
            if alias != profile["userName"]:
                sdk.detectionlists.remove_user_cloud_alias(user_id, alias)
        sdk.detectionlists.add_user_cloud_alias(user_id, cloud_alias)


def _update_risk_tags(sdk, username, risk_tag):
    if risk_tag:
        add_risk_tags(sdk, username, risk_tag)


def _update_notes(sdk, user_id, notes):
    if notes:
        sdk.detectionlists.update_user_notes(user_id, notes)


def add_risk_tags(sdk, username, risk_tag):
    risk_tag = handle_list_args(risk_tag)
    user_id = get_user_id(sdk, username)
    sdk.detectionlists.add_user_risk_tags(user_id, risk_tag)


def remove_risk_tags(sdk, username, risk_tag):
    risk_tag = handle_list_args(risk_tag)
    user_id = get_user_id(sdk, username)
    sdk.detectionlists.remove_user_risk_tags(user_id, risk_tag)


def handle_list_args(list_arg):
    """Converts str args to a list. Useful for `bulk` commands which don't use click's argument
    parsing but instead pass in values from files, such as in the form "item1 item2"."""
    if isinstance(list_arg, str):
        return list_arg.split()
    return list_arg



import click
from py42.clients.trustedactivities import TrustedActivityType

from code42cli.bulk import generate_template_cmd_factory
from code42cli.bulk import run_bulk_process
from code42cli.click_ext.groups import OrderedGroup
from code42cli.errors import Code42CLIError
from code42cli.file_readers import read_csv_arg
from code42cli.options import format_option
from code42cli.options import sdk_options
from code42cli.output_formats import OutputFormatter

resource_id_arg = click.argument("resource-id", type=int)
type_option = click.option(
    "--type",
    help=f"Type of trusted activity. Valid types include {', '.join(TrustedActivityType.choices())}.",
    type=click.Choice(TrustedActivityType.choices()),
)
value_option = click.option(
    "--value",
    help="The value of the trusted activity, such as the domain or Slack workspace name.",
)
description_option = click.option(
    "--description", help="The description of the trusted activity."
)


def _get_trust_header():
    return {
        "resourceId": "Resource Id",
        "type": "Type",
        "value": "Value",
        "description": "Description",
        "updatedAt": "Last Update Time",
        "updatedByUsername": "Last Updated By (Username)",
        "updatedByUserUid": "Last updated By (UserUID)",
    }


@click.group(cls=OrderedGroup)
@sdk_options(hidden=True)
def trusted_activities(state):
    """Manage trusted activities and resources."""
    pass


@trusted_activities.command()
@click.argument("type", type=click.Choice(TrustedActivityType.choices()))
@click.argument("value")
@description_option
@sdk_options()
def create(state, type, value, description):
    """Create a trusted activity.

    VALUE is the name of the domain or Slack workspace.
    """
    state.sdk.trustedactivities.create(
        type,
        value,
        description=description,
    )


@trusted_activities.command()
@resource_id_arg
@value_option
@description_option
@sdk_options()
def update(state, resource_id, value, description):
    """Update a trusted activity.  Requires the activity's resource ID."""
    state.sdk.trustedactivities.update(
        resource_id,
        value=value,
        description=description,
    )


@trusted_activities.command()
@resource_id_arg
@sdk_options()
def remove(state, resource_id):
    """Remove a trusted activity.  Requires the activity's resource ID."""
    state.sdk.trustedactivities.delete(resource_id)


@trusted_activities.command("list")
@click.option("--type", type=click.Choice(TrustedActivityType.choices()))
@format_option
@sdk_options()
def _list(state, type, format):
    """List all trusted activities."""
    pages = state.sdk.trustedactivities.get_all(type=type)
    formatter = OutputFormatter(format, _get_trust_header())
    trusted_resources = [
        resource for page in pages for resource in page["trustResources"]
    ]
    if trusted_resources:
        formatter.echo_formatted_list(trusted_resources)
    else:
        click.echo("No trusted activities found.")


@trusted_activities.group(cls=OrderedGroup)
@sdk_options(hidden=True)
def bulk(state):
    """Tools for executing bulk trusted activity actions."""
    pass


TRUST_CREATE_HEADERS = [
    "type",
    "value",
    "description",
]
TRUST_UPDATE_HEADERS = [
    "resource_id",
    "value",
    "description",
]
TRUST_REMOVE_HEADERS = [
    "resource_id",
]

trusted_activities_generate_template = generate_template_cmd_factory(
    group_name="trusted_activities",
    commands_dict={
        "create": TRUST_CREATE_HEADERS,
        "update": TRUST_UPDATE_HEADERS,
        "remove": TRUST_REMOVE_HEADERS,
    },
    help_message="Generate the CSV template needed for bulk trusted-activities commands",
)
bulk.add_command(trusted_activities_generate_template)


@bulk.command(
    name="create",
    help="Bulk create trusted activities using a CSV file with "
    f"format: {','.join(TRUST_CREATE_HEADERS)}.\b\n\n"
    f"Available `type` values are: {'|'.join(TrustedActivityType.choices())}",
)
@read_csv_arg(headers=TRUST_CREATE_HEADERS)
@sdk_options()
def bulk_create(state, csv_rows):
    """Bulk create trusted activities."""
    sdk = state.sdk

    def handle_row(type, value, description):
        if type not in TrustedActivityType.choices():
            message = f"Invalid type {type}, valid types include {', '.join(TrustedActivityType.choices())}."
            raise Code42CLIError(message)
        if type is None:
            message = "'type' is a required field to create a trusted activity."
            raise Code42CLIError(message)
        if value is None:
            message = "'value' is a required field to create a trusted activity."
            raise Code42CLIError(message)
        sdk.trustedactivities.create(type, value, description)

    run_bulk_process(
        handle_row,
        csv_rows,
        progress_label="Creating trusting activities:",
    )


@bulk.command(
    name="update",
    help="Bulk update trusted activities using a CSV file with "
    f"format: {','.join(TRUST_UPDATE_HEADERS)}.",
)
@read_csv_arg(headers=TRUST_UPDATE_HEADERS)
@sdk_options()
def bulk_update(state, csv_rows):
    """Bulk update trusted activities."""
    sdk = state.sdk

    def handle_row(resource_id, value, description):
        if resource_id is None:
            message = "'resource_id' is a required field to update a trusted activity."
            raise Code42CLIError(message)
        _check_resource_id_type(resource_id)
        sdk.trustedactivities.update(resource_id, value, description)

    run_bulk_process(
        handle_row, csv_rows, progress_label="Updating trusted activities:"
    )


@bulk.command(
    name="remove",
    help="Bulk remove trusted activities using a CSV file with "
    f"format: {','.join(TRUST_REMOVE_HEADERS)}.",
)
@read_csv_arg(headers=TRUST_REMOVE_HEADERS)
@sdk_options()
def bulk_remove(state, csv_rows):
    """Bulk remove trusted activities."""
    sdk = state.sdk

    def handle_row(resource_id):
        if resource_id is None:
            message = "'resource_id' is a required field to remove a trusted activity."
            raise Code42CLIError(message)
        _check_resource_id_type(resource_id)
        sdk.trustedactivities.delete(resource_id)

    run_bulk_process(
        handle_row,
        csv_rows,
        progress_label="Removing trusted activities:",
    )


def _check_resource_id_type(resource_id):
    def raise_error(resource_id):
        message = f"Invalid resource ID {resource_id}.  Must be an integer."
        raise Code42CLIError(message)

    try:
        if not float(resource_id).is_integer():
            raise_error(resource_id)
    except ValueError:
        raise_error(resource_id)



import click


def incompatible_with(incompatible_opts):
    """Factory for creating custom `click.Option` subclasses that enforce incompatibility with the
    option strings passed to this function.
    """

    if isinstance(incompatible_opts, str):
        incompatible_opts = [incompatible_opts]

    class IncompatibleOption(click.Option):
        def __init__(self, *args, **kwargs):
            super().__init__(*args, **kwargs)

        def handle_parse_result(self, ctx, opts, args):
            # if None it means we're in autocomplete mode and don't want to validate
            if ctx.obj is not None:
                found_incompatible = ", ".join(
                    [
                        f"--{opt.replace('_', '-')}"
                        for opt in opts
                        if opt in incompatible_opts
                    ]
                )
                if self.name in opts and found_incompatible:
                    name = self.name.replace("_", "-")
                    raise click.BadOptionUsage(
                        option_name=self.name,
                        message=f"--{name} can't be used with: {found_incompatible}",
                    )
            return super().handle_parse_result(ctx, opts, args)

    return IncompatibleOption



import json
from functools import lru_cache
from pprint import pformat

import click
from click import echo

from code42cli.bulk import generate_template_cmd_factory
from code42cli.bulk import run_bulk_process
from code42cli.click_ext.groups import OrderedGroup
from code42cli.cmds.shared import get_user_id
from code42cli.errors import UserNotInLegalHoldError
from code42cli.file_readers import read_csv_arg
from code42cli.options import format_option
from code42cli.options import sdk_options
from code42cli.options import set_begin_default_dict
from code42cli.options import set_end_default_dict
from code42cli.output_formats import OutputFormat
from code42cli.output_formats import OutputFormatter
from code42cli.util import format_string_list_to_columns


_MATTER_KEYS_MAP = {
    "legalHoldUid": "Matter ID",
    "name": "Name",
    "description": "Description",
    "creator_username": "Creator",
    "creationDate": "Creation Date",
}
_EVENT_KEYS_MAP = {
    "eventUid": "Event ID",
    "eventType": "Event Type",
    "eventDate": "Event Date",
    "legalHoldUid": "Legal Hold ID",
    "actorUsername": "Actor Username",
    "custodianUsername": "Custodian Username",
}
LEGAL_HOLD_KEYWORD = "legal hold events"
LEGAL_HOLD_EVENT_TYPES = [
    "MembershipCreated",
    "MembershipReactivated",
    "MembershipDeactivated",
    "HoldCreated",
    "HoldDeactivated",
    "HoldReactivated",
    "Restore",
]
BEGIN_DATE_DICT = set_begin_default_dict(LEGAL_HOLD_KEYWORD)
END_DATE_DICT = set_end_default_dict(LEGAL_HOLD_KEYWORD)


@click.group(cls=OrderedGroup)
@sdk_options(hidden=True)
def legal_hold(state):
    """Add and remove custodians from legal hold matters."""


def matter_id_option(required, help):
    return click.option("-m", "--matter-id", required=required, type=str, help=help)


user_id_option = click.option(
    "-u",
    "--username",
    required=True,
    type=str,
    help="The username of the custodian to add to the matter.",
)


@legal_hold.command()
@matter_id_option(
    True,
    "Identification number of the legal hold matter the custodian will be added to.",
)
@user_id_option
@sdk_options()
def add_user(state, matter_id, username):
    """Add a custodian to a legal hold matter."""
    _add_user_to_legal_hold(state.sdk, matter_id, username)


@legal_hold.command()
@matter_id_option(
    True,
    "Identification number of the legal hold matter the custodian will be removed from.",
)
@user_id_option
@sdk_options()
def remove_user(state, matter_id, username):
    """Release a custodian from a legal hold matter."""
    _remove_user_from_legal_hold(state.sdk, matter_id, username)


@legal_hold.command("list")
@format_option
@sdk_options()
def _list(state, format=None):
    """Fetch existing legal hold matters."""
    formatter = OutputFormatter(format, _MATTER_KEYS_MAP)
    matters = _get_all_active_matters(state.sdk)
    if matters:
        formatter.echo_formatted_list(matters)


@legal_hold.command()
@click.argument("matter-id")
@click.option(
    "--include-inactive",
    is_flag=True,
    help="View all custodians associated with the legal hold matter, "
    "including inactive custodians.",
)
@click.option(
    "--include-policy",
    is_flag=True,
    help="View details of the preservation policy associated with the legal hold matter.",
)
@sdk_options()
def show(state, matter_id, include_inactive=False, include_policy=False):
    """Display details of a given legal hold matter."""
    matter = _check_matter_is_accessible(state.sdk, matter_id)
    matter["creator_username"] = matter["creator"]["username"]
    matter = json.loads(matter.text)

    # if `active` is None then all matters (whether active or inactive) are returned. True returns
    # only those that are active.
    active = None if include_inactive else True
    memberships = _get_legal_hold_memberships_for_matter(
        state.sdk, matter_id, active=active
    )
    active_usernames = [
        member["user"]["username"] for member in memberships if member["active"]
    ]
    inactive_usernames = [
        member["user"]["username"] for member in memberships if not member["active"]
    ]

    formatter = OutputFormatter(OutputFormat.TABLE, _MATTER_KEYS_MAP)
    formatter.echo_formatted_list([matter])
    _print_matter_members(active_usernames, member_type="active")

    if include_inactive:
        _print_matter_members(inactive_usernames, member_type="inactive")

    if include_policy:
        _get_and_print_preservation_policy(state.sdk, matter["holdPolicyUid"])
        echo("")


@legal_hold.command()
@matter_id_option(False, "Filter results by legal hold UID.")
@click.option(
    "--event-type",
    type=click.Choice(LEGAL_HOLD_EVENT_TYPES),
    help="Filter results by event types.",
)
@click.option("--begin", **BEGIN_DATE_DICT)
@click.option("--end", **END_DATE_DICT)
@format_option
@sdk_options()
def search_events(state, matter_id, event_type, begin, end, format):
    """Tools for getting legal hold event data."""
    formatter = OutputFormatter(format, _EVENT_KEYS_MAP)
    events = _get_all_events(state.sdk, matter_id, begin, end)
    if event_type:
        events = [event for event in events if event["eventType"] == event_type]
    if events:
        formatter.echo_formatted_list(events)
    else:
        click.echo("No results found.")


@legal_hold.group(cls=OrderedGroup)
@sdk_options(hidden=True)
def bulk(state):
    """Tools for executing bulk legal hold actions."""
    pass


LEGAL_HOLD_CSV_HEADERS = ["matter_id", "username"]


legal_hold_generate_template = generate_template_cmd_factory(
    group_name="legal_hold",
    commands_dict={"add": LEGAL_HOLD_CSV_HEADERS, "remove": LEGAL_HOLD_CSV_HEADERS},
)
bulk.add_command(legal_hold_generate_template)


@bulk.command(
    name="add",
    help=f"Bulk add custodians to legal hold matters using a CSV file. "
    f"CSV file format: {','.join(LEGAL_HOLD_CSV_HEADERS)}",
)
@read_csv_arg(headers=LEGAL_HOLD_CSV_HEADERS)
@sdk_options()
def bulk_add(state, csv_rows):
    sdk = state.sdk

    def handle_row(matter_id, username):
        _add_user_to_legal_hold(sdk, matter_id, username)

    run_bulk_process(handle_row, csv_rows, progress_label="Adding users to legal hold:")


@bulk.command(
    help=f"Bulk release custodians from legal hold matters using a CSV file. "
    f"CSV file format: {','.join(LEGAL_HOLD_CSV_HEADERS)}"
)
@read_csv_arg(headers=LEGAL_HOLD_CSV_HEADERS)
@sdk_options()
def remove(state, csv_rows):
    sdk = state.sdk

    def handle_row(matter_id, username):
        _remove_user_from_legal_hold(sdk, matter_id, username)

    run_bulk_process(
        handle_row, csv_rows, progress_label="Removing users from legal hold:"
    )


def _add_user_to_legal_hold(sdk, matter_id, username):
    user_id = get_user_id(sdk, username)
    _check_matter_is_accessible(sdk, matter_id)
    sdk.legalhold.add_to_matter(user_id, matter_id)


def _remove_user_from_legal_hold(sdk, matter_id, username):
    _check_matter_is_accessible(sdk, matter_id)
    membership_id = _get_legal_hold_membership_id_for_user_and_matter(
        sdk, username, matter_id
    )
    sdk.legalhold.remove_from_matter(membership_id)


def _get_and_print_preservation_policy(sdk, policy_uid):
    preservation_policy = sdk.legalhold.get_policy_by_uid(policy_uid)
    echo("\nPreservation Policy:\n")
    echo(pformat(json.loads(preservation_policy.text)))


def _get_legal_hold_membership_id_for_user_and_matter(sdk, username, matter_id):
    user_id = get_user_id(sdk, username)
    memberships = _get_legal_hold_memberships_for_matter(sdk, matter_id, active=True)
    for member in memberships:
        if member["user"]["userUid"] == user_id:
            return member["legalHoldMembershipUid"]
    raise UserNotInLegalHoldError(username, matter_id)


def _get_legal_hold_memberships_for_matter(sdk, matter_id, active=True):
    memberships_generator = sdk.legalhold.get_all_matter_custodians(
        legal_hold_uid=matter_id, active=active
    )
    memberships = [
        member
        for page in memberships_generator
        for member in page["legalHoldMemberships"]
    ]
    return memberships


def _get_all_active_matters(sdk):
    matters_generator = sdk.legalhold.get_all_matters()
    matters = [
        matter
        for page in matters_generator
        for matter in page["legalHolds"]
        if matter["active"]
    ]
    for matter in matters:
        matter["creator_username"] = matter["creator"]["username"]
    return matters


def _get_all_events(sdk, legal_hold_uid, begin_date, end_date):
    events_generator = sdk.legalhold.get_all_events(
        legal_hold_uid, begin_date, end_date
    )
    events = [event for page in events_generator for event in page["legalHoldEvents"]]
    return events


def _print_matter_members(username_list, member_type="active"):
    if username_list:
        echo(f"\n{member_type.capitalize()} matter members:\n")
        format_string_list_to_columns(username_list)
    else:
        echo(f"No {member_type} matter members.\n")


@lru_cache(maxsize=None)
def _check_matter_is_accessible(sdk, matter_id):
    return sdk.legalhold.get_matter_by_uid(matter_id)



import logging
import os
import traceback
from logging.handlers import RotatingFileHandler
from threading import Lock

from code42cli.enums import FileEventsOutputFormat
from code42cli.logger.formatters import FileEventDictToCEFFormatter
from code42cli.logger.formatters import FileEventDictToJSONFormatter
from code42cli.logger.formatters import FileEventDictToRawJSONFormatter
from code42cli.logger.handlers import NoPrioritySysLogHandler
from code42cli.util import get_url_parts
from code42cli.util import get_user_project_path

# prevent loggers from printing stacks to stderr if a pipe is broken
logging.raiseExceptions = False

logger_deps_lock = Lock()
ERROR_LOG_FILE_NAME = "code42_errors.log"


def _get_formatter(output_format):
    if output_format == FileEventsOutputFormat.JSON:
        return FileEventDictToJSONFormatter()
    elif output_format == FileEventsOutputFormat.CEF:
        return FileEventDictToCEFFormatter()
    else:
        return FileEventDictToRawJSONFormatter()


def _init_logger(logger, handler, output_format):
    formatter = _get_formatter(output_format)
    logger.setLevel(logging.INFO)
    return add_handler_to_logger(logger, handler, formatter)


def get_logger_for_server(hostname, protocol, output_format, certs):
    """Gets the logger that sends logs to a server for the given format.

    Args:
        hostname: The hostname of the server. It may include the port.
        protocol: The transfer protocol for sending logs.
        output_format: CEF, JSON, or RAW_JSON. Each type results in a different logger instance.
        certs: Use for passing SSL/TLS certificates when connecting to the server.
    """
    logger = logging.getLogger(f"code42_syslog_{output_format.lower()}")
    if logger_has_handlers(logger):
        return logger

    with logger_deps_lock:
        url_parts = get_url_parts(hostname)
        hostname = url_parts[0]
        port = url_parts[1] or 514
        if not logger_has_handlers(logger):
            handler = NoPrioritySysLogHandler(hostname, port, protocol, certs)
            handler.connect_socket()
            return _init_logger(logger, handler, output_format)
    return logger


def _get_standard_formatter():
    return logging.Formatter("%(message)s")


def _get_error_log_path():
    log_path = get_user_project_path("log")
    return os.path.join(log_path, ERROR_LOG_FILE_NAME)


def _create_error_file_handler():
    log_path = _get_error_log_path()
    return RotatingFileHandler(
        log_path, maxBytes=250000000, encoding="utf-8", delay=True
    )


def add_handler_to_logger(logger, handler, formatter):
    handler.setFormatter(formatter)
    logger.addHandler(handler)
    return logger


def logger_has_handlers(logger):
    return len(logger.handlers)


def _get_error_file_logger():
    """Gets the logger where raw exceptions are logged."""
    logger = logging.getLogger("code42_error_logger")
    if logger_has_handlers(logger):
        return logger

    with logger_deps_lock:
        if not logger_has_handlers(logger):
            formatter = _create_formatter_for_error_file()
            handler = _create_error_file_handler()
            return add_handler_to_logger(logger, handler, formatter)
    return logger


def get_view_error_details_message():
    """Returns the error message that is printed when errors occur."""
    path = _get_error_log_path()
    return f"View details in {path}"


def _create_formatter_for_error_file():
    return logging.Formatter("%(asctime)s %(message)s")


class CliLogger:
    def __init__(self):
        self._logger = _get_error_file_logger()

    def log_error(self, err):
        message = str(err) if err else None
        if message:
            self._logger.error(message)

    def log_verbose_error(self, invocation_str=None, http_request=None):
        """For logging traces, invocation strs, and request parameters during exceptions to the
        error log file."""
        prefix = (
            "Exception occurred."
            if not invocation_str
            else f"Exception occurred from input: '{invocation_str}'."
        )
        message = f"{prefix}. See error below."
        self.log_error(message)
        self.log_error(traceback.format_exc())
        if http_request:
            self.log_error(f"Request parameters: {http_request.body}")


def get_main_cli_logger():
    return CliLogger()



import json
from datetime import datetime
from logging import Formatter

from code42cli.maps import CEF_CUSTOM_FIELD_NAME_MAP
from code42cli.maps import FILE_EVENT_TO_SIGNATURE_ID_MAP
from code42cli.maps import JSON_TO_CEF_MAP

CEF_TEMPLATE = (
    "CEF:0|Code42|{productName}|1|{signatureID}|{eventName}|{severity}|{extension}"
)
CEF_TIMESTAMP_FIELDS = ["end", "fileCreateTime", "fileModificationTime", "rt"]


class FileEventDictToCEFFormatter(Formatter):
    """Formats file event dicts into CEF format. Attach to a logger via `setFormatter` to use.
    Args:
        default_product_name: The default value to use in the product name segment of the CEF message.
        default_severity_level: The default integer between 1 and 10 to assign to the severity segment of the CEF message.
    """

    def __init__(
        self,
        default_product_name="Advanced Exfiltration Detection",
        default_severity_level="5",
    ):
        super().__init__()
        self._default_product_name = default_product_name
        self._default_severity_level = default_severity_level

    def format(self, record):
        """
        Args:
            record (LogRecord): `record.msg` must be a `dict`.
        """
        file_event_dict = record.msg
        # security events must convert to file event dict format before calling this.
        ext, evt, sig_id = map_event_to_cef(file_event_dict)
        cef_log = CEF_TEMPLATE.format(
            productName=self._default_product_name,
            signatureID=sig_id,
            eventName=evt,
            severity=self._default_severity_level,
            extension=ext,
        )
        return cef_log


class FileEventDictToJSONFormatter(Formatter):
    """Formats file event dicts into JSON format. Attach to a logger via `setFormatter` to use.
    Items in the dictionary whose values are `None`, empty string, or empty lists will be excluded
    from the JSON conversion.
    """

    def format(self, record):
        """
        Args:
            record (LogRecord): `record.msg` must be a `dict`.
        """
        file_event_dict = record.msg
        file_event_dict = {
            key: file_event_dict[key]
            for key in file_event_dict
            if file_event_dict[key] or file_event_dict[key] == 0
        }
        return json.dumps(file_event_dict)


class FileEventDictToRawJSONFormatter(Formatter):
    """Formats file event dicts into JSON format. Attach to a logger via `setFormatter` to use."""

    def format(self, record):
        return json.dumps(record.msg)


def _format_cef_kvp(cef_field_key, cef_field_value):
    if cef_field_key + "Label" in CEF_CUSTOM_FIELD_NAME_MAP:
        return _format_custom_cef_kvp(cef_field_key, cef_field_value)

    cef_field_value = _handle_nested_json_fields(cef_field_key, cef_field_value)
    if isinstance(cef_field_value, list):
        cef_field_value = _convert_list_to_csv(cef_field_value)
    elif cef_field_key in CEF_TIMESTAMP_FIELDS:
        cef_field_value = convert_file_event_timestamp_to_cef_timestamp(cef_field_value)
    return f"{cef_field_key}={cef_field_value}"


def _handle_nested_json_fields(cef_field_key, cef_field_value):
    result = []
    if cef_field_key == "duser":
        result = [
            item["cloudUsername"] for item in cef_field_value if type(item) is dict
        ]

    return result or cef_field_value


def _format_custom_cef_kvp(custom_cef_field_key, custom_cef_field_value):
    custom_cef_label_key = f"{custom_cef_field_key}Label"
    custom_cef_label_value = CEF_CUSTOM_FIELD_NAME_MAP[custom_cef_label_key]
    return (
        f"{custom_cef_field_key}={custom_cef_field_value} "
        f"{custom_cef_label_key}={custom_cef_label_value}"
    )


def _convert_list_to_csv(_list):
    value = ",".join([val for val in _list])
    return value


def convert_file_event_timestamp_to_cef_timestamp(timestamp_value):
    try:
        _datetime = datetime.strptime(timestamp_value, "%Y-%m-%dT%H:%M:%S.%fZ")
    except ValueError:
        _datetime = datetime.strptime(timestamp_value, "%Y-%m-%dT%H:%M:%SZ")
    value = f"{_datetime_to_ms_since_epoch(_datetime):.0f}"
    return value


def _datetime_to_ms_since_epoch(_datetime):
    epoch = datetime.utcfromtimestamp(0)
    total_seconds = (_datetime - epoch).total_seconds()
    # total_seconds will be in decimals (millisecond precision)
    return total_seconds * 1000


def map_event_to_cef(event):
    kvp_list = {
        JSON_TO_CEF_MAP[key]: event[key]
        for key in event
        if key in JSON_TO_CEF_MAP and (event[key] is not None and event[key] != [])
    }
    extension = " ".join(_format_cef_kvp(key, kvp_list[key]) for key in kvp_list)
    event_name = event.get("eventType", "UNKNOWN")
    signature_id = FILE_EVENT_TO_SIGNATURE_ID_MAP.get(event_name, "C42000")
    return extension, event_name, signature_id



import click
from click._compat import get_text_stderr

from code42cli.logger import get_view_error_details_message

ERRORED = False


class Code42CLIError(click.ClickException):
    """Base CLI exception. The `message` param automatically gets logged to error file and printed
    to stderr in red text. If `help` param is provided, it will also be printed to stderr after the
    message but not logged to file.
    """

    def __init__(self, message, help=None):
        self.help = help
        super().__init__(message)

    def show(self, file=None):
        """Override default `show` to print CLI errors in red text."""
        if file is None:
            file = get_text_stderr()
        click.secho(f"Error: {self.format_message()}", file=file, fg="red")
        if self.help:
            click.echo(self.help, err=True)


class LoggedCLIError(Code42CLIError):
    """Exception to be raised when wanting to point users to error logs for error details.

    If `message` param is provided it will be printed to screen along with message on where to
    find error details in the log.
    """

    def __init__(self, message=None):
        self.message = message
        super().__init__(message)

    def format_message(self):
        locations_message = get_view_error_details_message()
        return (
            f"{self.message}\n{locations_message}"
            if self.message
            else locations_message
        )


class UserDoesNotExistError(Code42CLIError):
    """An error to represent a username that is not in our system. The CLI shows this error when
    the user tries to add or remove a user that does not exist. This error is not shown during
    bulk add or remove."""

    def __init__(self, username):
        super().__init__(
            f"User '{username}' does not exist or you do not have permission to view them."
        )


class UserNotInLegalHoldError(Code42CLIError):
    def __init__(self, username, matter_id):
        super().__init__(
            f"User '{username}' is not an active member of legal hold matter '{matter_id}'."
        )



from py42.choices import Choices


class JsonOutputFormat(Choices):
    JSON = "JSON"
    RAW = "RAW-JSON"

    def __iter__(self):
        return iter([self.JSON, self.RAW])


class OutputFormat(JsonOutputFormat):
    TABLE = "TABLE"
    CSV = "CSV"

    def __iter__(self):
        return iter([self.TABLE, self.CSV, self.JSON, self.RAW])


class SendToFileEventsOutputFormat(JsonOutputFormat):
    CEF = "CEF"

    def __iter__(self):
        return iter([self.CEF, self.JSON, self.RAW])


class FileEventsOutputFormat(OutputFormat):
    CEF = "CEF"

    def __iter__(self):
        return iter([self.TABLE, self.CSV, self.JSON, self.RAW, self.CEF])



import click
from py42.services.detectionlists.departing_employee import DepartingEmployeeFilters

from code42cli.bulk import generate_template_cmd_factory
from code42cli.bulk import run_bulk_process
from code42cli.click_ext.groups import OrderedGroup
from code42cli.cmds.detectionlists import ALL_FILTER
from code42cli.cmds.detectionlists import get_choices
from code42cli.cmds.detectionlists import handle_filter_choice
from code42cli.cmds.detectionlists import list_employees
from code42cli.cmds.detectionlists import update_user
from code42cli.cmds.detectionlists.options import cloud_alias_option
from code42cli.cmds.detectionlists.options import notes_option
from code42cli.cmds.detectionlists.options import username_arg
from code42cli.cmds.shared import get_user_id
from code42cli.errors import Code42CLIError
from code42cli.file_readers import read_csv_arg
from code42cli.options import format_option
from code42cli.options import sdk_options
from code42cli.util import deprecation_warning


def _get_filter_choices():
    filters = DepartingEmployeeFilters.choices()
    return get_choices(filters)


DEPRECATION_TEXT = "(DEPRECATED): Use `code42 watchlists` commands instead."

DATE_FORMAT = "%Y-%m-%d"
filter_option = click.option(
    "--filter",
    help=f"Departing employee filter options. Defaults to {ALL_FILTER}.",
    type=click.Choice(_get_filter_choices()),
    default=ALL_FILTER,
    callback=lambda ctx, param, arg: handle_filter_choice(arg),
)


@click.group(
    cls=OrderedGroup,
    help=f"{DEPRECATION_TEXT}\n\nAdd and remove employees from the Departing Employees detection list.",
)
@sdk_options(hidden=True)
def departing_employee(state):
    pass


@departing_employee.command(
    "list",
    help=f"{DEPRECATION_TEXT}\n\nLists the users on the Departing Employees list.",
)
@sdk_options()
@format_option
@filter_option
def _list(state, format, filter):
    deprecation_warning(DEPRECATION_TEXT)
    employee_generator = _get_departing_employees(state.sdk, filter)
    list_employees(
        employee_generator,
        format,
        {"departureDate": "Departure Date"},
    )


@departing_employee.command(
    help=f"{DEPRECATION_TEXT}\n\nAdd a user to the Departing Employees detection list."
)
@username_arg
@click.option(
    "--departure-date",
    help="The date the employee is departing. Format: yyyy-MM-dd.",
    type=click.DateTime(formats=[DATE_FORMAT]),
)
@cloud_alias_option
@notes_option
@sdk_options()
def add(state, username, cloud_alias, departure_date, notes):

    deprecation_warning(DEPRECATION_TEXT)
    _add_departing_employee(state.sdk, username, cloud_alias, departure_date, notes)


@departing_employee.command(
    help=f"{DEPRECATION_TEXT}\n\nRemove a user from the Departing Employees detection list."
)
@username_arg
@sdk_options()
def remove(state, username):
    deprecation_warning(DEPRECATION_TEXT)
    _remove_departing_employee(state.sdk, username)


@departing_employee.group(
    cls=OrderedGroup,
    help=f"{DEPRECATION_TEXT}\n\nTools for executing bulk departing employee actions.",
)
@sdk_options(hidden=True)
def bulk(state):
    pass


DEPARTING_EMPLOYEE_CSV_HEADERS = ["username", "cloud_alias", "departure_date", "notes"]

REMOVE_EMPLOYEE_HEADERS = ["username"]

departing_employee_generate_template = generate_template_cmd_factory(
    group_name="departing_employee",
    commands_dict={
        "add": DEPARTING_EMPLOYEE_CSV_HEADERS,
        "remove": REMOVE_EMPLOYEE_HEADERS,
    },
)
bulk.add_command(departing_employee_generate_template)


@bulk.command(
    name="add",
    help=f"{DEPRECATION_TEXT}\n\nBulk add users to the departing employees detection list using "
    f"a CSV file with format: {','.join(DEPARTING_EMPLOYEE_CSV_HEADERS)}.",
)
@read_csv_arg(headers=DEPARTING_EMPLOYEE_CSV_HEADERS)
@sdk_options()
def bulk_add(state, csv_rows):
    deprecation_warning(DEPRECATION_TEXT)
    sdk = state.sdk  # Force initialization of py42 to only happen once.

    def handle_row(username, cloud_alias, departure_date, notes):
        if departure_date:
            try:
                departure_date = click.DateTime(formats=[DATE_FORMAT]).convert(
                    departure_date, None, None
                )
            except click.exceptions.BadParameter:
                message = (
                    f"Invalid date {departure_date}, valid date format {DATE_FORMAT}."
                )
                raise Code42CLIError(message)
        _add_departing_employee(sdk, username, cloud_alias, departure_date, notes)

    run_bulk_process(
        handle_row,
        csv_rows,
        progress_label="Adding users to the Departing Employees detection list:",
    )


@bulk.command(
    name="remove",
    help=f"{DEPRECATION_TEXT}\n\nBulk remove users from the departing employees detection list "
    f"using a CSV file with format {','.join(REMOVE_EMPLOYEE_HEADERS)}.",
)
@read_csv_arg(headers=REMOVE_EMPLOYEE_HEADERS)
@sdk_options()
def bulk_remove(state, csv_rows):
    deprecation_warning(DEPRECATION_TEXT)
    sdk = state.sdk

    def handle_row(username):
        _remove_departing_employee(sdk, username)

    run_bulk_process(
        handle_row,
        csv_rows,
        progress_label="Removing users from the Departing Employees detection list:",
    )


def _get_departing_employees(sdk, filter):
    return sdk.detectionlists.departing_employee.get_all(filter)


def _add_departing_employee(sdk, username, cloud_alias, departure_date, notes):
    if departure_date:
        departure_date = departure_date.strftime(DATE_FORMAT)
    user_id = get_user_id(sdk, username)
    sdk.detectionlists.departing_employee.add(user_id, departure_date)
    update_user(sdk, username, cloud_alias=cloud_alias, notes=notes)


def _remove_departing_employee(sdk, username):
    user_id = get_user_id(sdk, username)
    sdk.detectionlists.departing_employee.remove(user_id)



import click

from code42cli import BANNER
from code42cli.options import sdk_options


@click.command()
@sdk_options()
def shell(state):
    """Open an IPython shell with py42 initialized as `sdk`."""
    import IPython

    IPython.embed(colors="Neutral", banner1=BANNER, user_ns={"sdk": state.sdk})
